from threading import Event, Timer, Thread
from queue import Queue
import os
import requests
import json
import glob
import re
import abc
import math
import time
import datetime
import textwrap
import pandas as pd

import sseclient

# Defaults
VERSION = "1.0"
OUTPUT_VERSION = 1

#=====================================================================
# Abstract crawler class
#=====================================================================

class Crawler:
    """ Abstract class representing a crawler
    """

    url = ""
    query = ""
    path = ""
    writes = 0
    exceptions = 0
    warnings = 0
    compress = False
    stop_request = Event()
    running = Event()
    message_webhook_url = None
    messages = Queue()
    authkey = ""
    DEBUG = False

    def __init__(self, url, path, authkey, version=VERSION, output_version=OUTPUT_VERSION):
        """Creates a Crawler object

        Args:
            url (str): The endpoint URL where the crawler fetches data from
            path (str): The local path to save crawled data in
        """
        self.url = url
        self.path = path
        self.authkey = authkey
        self.VERSION = version
        self.OUTPUT_VERSION = output_version

        if not os.path.exists(self.path):
            os.makedirs(self.path)
        
        init_msg = "{name} (version {version}, output version {output_version}) initialized.".format(
            name = str(self),
            version = self.VERSION,
            output_version = self.OUTPUT_VERSION
        )
        self.log_status(init_msg, type="INIT")


    def __str__(self):
        return "mur-crawl@" + self.path

    def _attempt(self, func, max_attempts = 3, description="Operation"):
        """Makes several attempts to execute a function until it finishes without raising an exception.

        Args:
            func (function): The function to execute
            max_attempts (int, optional): Maximum number of attempts to make before raising the exception raised by func. Defaults to 3.
            description (str, optional): The description of the operation using for logging attempts. Defaults to "Operation".
        Returns: the value returned by func
        """

        attempt = 1
        while attempt <= max_attempts:
            try:
                #try to execute func
                return func()
            except Exception as e:
                #Check stop
                if self.stop_request.is_set():
                    self.stop()
                    self.running.clear()
                    raise Exception("Stop requested")
                
                #Retry?
                if attempt <= max_attempts:
                    wait_seconds = int(10*math.pow(3,attempt))
                    self.log_warning("{description} failed ({message}). Retrying in {wait_seconds}s...".format(description=description, message = str(e), wait_seconds=wait_seconds))
                    time.sleep(wait_seconds)
                    attempt += 1
                else:
                    self.log_exception(e)
                    self.log_warning("{description} failed.", description=description)
                    raise e

    def _query(self, query):
        headers = {'Content-Type': "text/xml"}
        if self.DEBUG:
            self.log_info(query)
        response = requests.post(self.url, headers=headers, data=query.encode("utf-8"))

        if not response.status_code == 200:
            self.log_warning("Query returned status code {statusCode}: {query}".format(statusCode=response.status_code, query = query))

        if self.DEBUG:
            self.log_info("Received: " + textwrap.shorten(response.text, width=1024))

        response = json.loads(response.text)
        return response

    @property
    def logfile(self):
        """Path of the csv file used for logging messages of this crawler

        Returns:
            str: Path to logfile
        """
        return os.path.join(self.path, "log.csv")

    def start(self):
        """Starts crawling
        """
        self.running.set()

    def stop(self):
        """Requests the crawler to stop crawling
        """
        self.stop_request.set()

    def send_message_to_webhook(self, title, message):
        """Sends a message using webhook

        Args:
            title (str): Message title
            message (str): Message text
        """
        if not self.message_webhook_url is None:
            try:
                data = {"value1": str(self),
                        "value2": title,
                        "value3": message}
                requests.post(self.message_webhook_url, data = data)
            except Exception as e:
                print("Failed to send message to webhook: " + str(e))

    def write_message_to_log(self, type, message):
        """Writes a message to the crawler's logfile

        Args:
            type (str): Message type (for example Exception, Warning)
            message (str): Message text
        """
        timestr = datetime.datetime.now().astimezone(tz=datetime.timezone.utc).isoformat(timespec='seconds').replace("+00:00","Z")
        data =  pd.DataFrame({"Time": [timestr], "Type": [type], "Message": [message]})
        #write to csv
        data.to_csv(self.logfile, index=False, mode='a', header=(not os.path.exists(self.logfile)))

    def log_exception(self, e):
        """Logs an exception that occoured

        Args:
            e (Exception): The exception to log
        """
        self.messages.put("EXCEPTION: {type} {message}".format(type=str(type(e)), message=str(e)))
        self.exceptions += 1

        self.send_message_to_webhook("Exception", str(e))
        self.write_message_to_log("Exception", str(e))
    
    def log_warning(self, msg):
        """Logs a warning message

        Args:
            msg (str): Warning message text
        """
        self.messages.put("WARNING: " + msg)
        self.warnings += 1
    
        self.send_message_to_webhook("Warning", msg)
        self.write_message_to_log("Warning", msg)

    def log_status(self, msg, type="Status"):
        """Logs a status message (normal operation).

        Args:
            msg (str): Status message
            type (str): Message type
        """
        self.write_message_to_log(type, msg)
        if self.DEBUG:
            self.messages.put("{type}: {msg}".format(type=type, msg=msg))

    def log_info(self, msg):
        """Logs an information message (for example data received). This is should be only used in when DEBUG is True.

        Args:
            msg (str): Info message
        """
        self.messages.put("INFO: " + msg)
    
    def join(self, wait_seconds = None):
        """Waits until the crawler has finished and returns True.

        Args:
            wait_seconds (float, optional): Return False if the crawler did not finish when this time [s] elapsed. Defaults to None (wait infinitely).

        Returns:
            bool: True if crawler finished, False if crawler did not finish but the waiting time has elapsed
        """
        while (wait_seconds is None or wait_seconds > 0) and self.running.is_set():
            time.sleep(1)
            wait_seconds -= 1

        if not self.running.is_set():
            return True
        else:
            return False

#=====================================================================
# Periodic crawler
#=====================================================================

class PeriodicCrawler(Crawler):
    """Abstract base class for a crawler fetching data periodically
    """
    interval = None
    timer = None

    def __init__(self, url, path, authkey, interval = 60, version=VERSION, output_version=OUTPUT_VERSION):
        """Creates a periodic crawler

        Args:
            url (str): The endpoint URL where the crawler fetches data from
            path (str): The local path to save crawled data in
            interval (int, optional): The interval to wait between fetching data in seconds. Defaults to 60.
        """
        super().__init__(url, path, authkey, version=version, output_version=output_version)
        self.interval = interval

        init_msg = "{name} (version {version}, output version {output_version}) initialized.".format(
            name = str(self),
            version = self.VERSION,
            output_version = self.OUTPUT_VERSION
        )
        self.log_status(init_msg, type="INIT")

    def start(self):
        """Starts the crawler to run in the bakground
        """
        super().start()

        def crawl_periodic():
            if not self.stop_request.is_set():
                try:
                    self.crawl()
                except Exception as e:
                    self.log_exception(e)

                #Schedule next crawl
                self.timer = Timer(self.interval, crawl_periodic)
                self.timer.start()
            else:
                self.running.clear()

        #Schedule first crawl
        self.timer = Timer(1, crawl_periodic)
        self.timer.start()
    
    def stop(self):
        super().stop()
        self.timer.cancel()
        self.running.clear()


#=====================================================================
# Stream crawler
#=====================================================================

class StreamCrawler(Crawler):
    """Abstract base class for a streaming crawler using Server-sent-events (SSE)
    """
    thread = None
    fileid = -1
    sseurl = None
    sse_stale = False
    client = None
    last_event = datetime.datetime.now()
    guard_events = None
    guard_connection = None
    connected = Event()
    inactivity_seconds = 8*60*60 #restart connection if no events received for 8 hours

    def __init__(self, url, path, authkey, version=VERSION, output_version=OUTPUT_VERSION):
        super().__init__(url, path, authkey, version=version, output_version=output_version)

    @property
    def fileprefix(self):
        return ""
    
    @property
    def statusfile(self):
        return os.path.join(self.path, self.fileprefix + str(self.fileid) + ".status")

    @property
    def csvfile(self):
        return os.path.join(self.path, self.fileprefix + str(self.fileid) + ".csv")

    def latest_fileid(self):
        """Get the latest fileid which exists in the output path

        Returns:
            int: The fileid or -1 if no file exists
        """

        #continue with latest file (csv with highest number)
        latest_fileid = -1
        csv_files = glob.glob(os.path.join(self.path, "*.csv*"))
        for filename in csv_files:
            m = re.match("[\s\S]*\/" + self.fileprefix + "([0-9]+).csv*$", filename)
            if m and int(m.group(1)) > latest_fileid:
                latest_fileid = int(m.group(1))
        return latest_fileid

    def _init_sse(self, query):
        """Initialize a new SSE (Server-Sent Events) connection using a query

        Args:
            query (str): The query to establish an SSE connection for

        Returns:
            sseclient.SSEClient: An SSEClient for the established connection
        """

        response = self._query(query)
        self.sseurl = response["RESPONSE"]["RESULT"][0]["INFO"]["SSEURL"]

        headers = {'Content-Type': "text/event-stream"}
        stream = requests.post(self.sseurl, stream=True, headers=headers)

        if not stream.status_code == 200:
            raise Exception("Returned status code {statusCode}: {url}".format(statusCode=stream.status_code, url = self.sseurl))

        return sseclient.SSEClient(stream)

    def _reconnect_sse(self, sseurl, lasteventid = None):
        """Reconnect to an SSE connection using an SSE url

        Args:
            sseurl (str): The SSE url to connect to
            lasteventid (bigint, optional): If specified, only events after this id will be streamed. Defaults to None.

        Returns:
            sseclient.SSEClient: An SSEClient for the established connection
        """
        self.sseurl = sseurl

        if not lasteventid is None:
            sseurl += "&lasteventid=" + str(lasteventid)
        
        headers = {'Content-Type': "text/event-stream"}
        stream = requests.post(sseurl, stream=True, headers=headers) #TODO: handle errors

        if not stream.status_code == 200:
            raise Exception("Returned status code {statusCode}: {url}".format(statusCode=stream.status_code, url = sseurl))

        return sseclient.SSEClient(stream)

    @abc.abstractmethod
    def _get_query(self):
        raise NotImplemented

    @abc.abstractmethod
    def _handle_events(self, data):
        raise NotImplemented
    
    def start(self, force_new_connection=False):
        # Stop if already running before restarting

        if not self.client is None:
            self.client.close()
            self.thread.join(timeout=60)
            self.client = None
       
        #Start guard to reconnect if no events have been received in long time
        def guard_events():
            if not self.stop_request.is_set():
                return
            
            if datetime.datetime.now() - self.last_event > datetime.timedelta(seconds=2.5*self.inactivity_seconds):
                self.log_exception(Exception("No events received in extremely long time though the connection seems up and running. A new connection will be established leading to a new csv file and possibly duplicate data."))
                self.sse_stale = True
                self.client.close() #closes current connection to trigger reconnect
            elif datetime.datetime.now() - self.last_event > datetime.timedelta(seconds=self.inactivity_seconds):
                self.log_warning("No events received in a while. Reconnecting...")
                self.client.close() #closes current connection to trigger reconnect
                
            self.guard_events = Timer(60, guard_events)
            self.guard_events.start()
        
        if not self.guard_events is None:
            self.guard_events.cancel()

        self.guard_events = Timer(60, guard_events)
        self.guard_events.start()

        def guard_connection():
            if not self.stop_request.is_set() and not self.connected.is_set():
                self.log_warning("No connection for more than 10 minutes. Please check the logfile for more information on the reason.")

        if not self.guard_connection is None:
            self.guard_connection.cancel()

        self.guard_connection = Timer(10*60, guard_events)
        self.guard_connection.start()

        #Look for previous connections and reconnect if possible
        if self.fileid == -1:
            self.fileid = self.latest_fileid()

        if not force_new_connection and os.path.isfile(self.statusfile):
            self.log_status("Reconnecting...")

            status = json.loads(open(self.statusfile, 'r').read())

            if not "version" in status or status["version"] < self.OUTPUT_VERSION:
                #Start new file if file format has changed
                self.log_warning("Output version change detected. Incoming events will be written to new csv file.")
                self.fileid = self.latest_fileid() + 1

            if "sseurl" in status and not status["sseurl"] is None:
                def reconnect():
                    self.client = self._reconnect_sse(status["sseurl"], status["lasteventid"])

                try:
                    self._attempt(reconnect, description="Reconnect SSE")
                except Exception as e:
                    self.client = None
            else:
                self.log_warning("Cannot reconnect due to missing sseurl")

        #Create new file and new connection if needed
        if self.client is None:
            self.log_status("Initializing new connection...")

            self.fileid = self.latest_fileid() + 1
            self.sseurl = None

            def init():
                self.client = self._init_sse(self._get_query())

            self._attempt(init, description="Init SSE", max_attempts=5)

        #Start stream in new thread
        def stream():
            reason = ""
            try:
                self.log_status("Streaming started", type="STREAM_START")
                self.connected.set()

                for event in self.client.events():
                    if self.stop_request.is_set():
                        self.stop()
                        self.running.clear()
                        return
                    
                    data = json.loads(event.data)
                    self._handle_events(data)
                    status = {"sseurl": self.sseurl, "lasteventid": event.id, "version": self.OUTPUT_VERSION, "lastevent": datetime.datetime.now().isoformat()}
                    with open(self.statusfile, 'w') as statusfile:
                        json.dump(status, statusfile)
                    self.last_event = datetime.datetime.now()
                    
            except Exception as e:
                reason = " (" + str(e) + ")"
            
            self.log_status("Streaming ended{reason}.".format(reason=reason), type="STREAM_END")
            self.connected.clear()
            self.guard_connection = Thread(target=guard_connection, daemon=True)
            self.start()

            #check stop
            if self.stop_request.is_set():
                self.stop()
                self.running.clear()
                return
            
            #connection ended, restart (force new connection if sse session is marked as stale)
            self.log_status("Reconnecting...".format(reason=reason))
            self.start(force_new_connection=self.sse_stale)

        self.thread = Thread(target=stream, daemon=True)
        super().start()
        self.thread.start()

    def stop(self):
        super().stop()
        if not self.client is None:
            self.client.close()
            self.client = None

    def join(self, wait_seconds = None):
        while (wait_seconds is None or wait_seconds > 0) and self.running.is_set():
            self.thread.join(timeout=1)
            wait_seconds -= 1

        if not self.running.is_set():
            return True
        else:
            return False
